//******************************************************************************
//
// File:    StatTest01.java
// Package: ---
// Unit:    Class StatTest01
//
// This Java source file is copyright (C) 2017 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This Java source file is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html.
//
//******************************************************************************

import edu.rit.io.InStream;
import edu.rit.io.Tabular;
import edu.rit.numeric.Histogram;
import edu.rit.pj2.Loop;
import edu.rit.pj2.Task;
import edu.rit.util.Random;
import edu.rit.util.RandomSubset;
import java.io.File;
import java.io.FileInputStream;

/**
 * Class StatTest01 is a multicore parallel main program that performs a
 * statistical test on the LC4 cipher.
 * <P>
 * Usage: <TT>java pj2 StatTest01 [-v] <I>dfreq</I> <I>Nlen</I> <I>Plen</I>
 * <I>trials</I> <I>reps</I> <I>seed</I></TT>
 * <P>
 * The program performs <I>reps</I> repetitions in parallel. For each
 * repetition, the program picks a random key and a random plaintext of length
 * <I>Plen</I> using the digram frequencies in the <I>dfreq</I> file. The
 * program does <I>trials</I> trials. For each trial, the program picks a random
 * nonce of length <I>Nlen</I>; computes the ciphertext from the key, nonce, and
 * plaintext; and tests whether the ciphertext letters and digrams for each
 * position obey a uniform distribution. However, if 36<SUP><I>Nlen</I></SUP>
 * &lt; <I>T</I>, the program tests all possible nonces.
 * <P>
 * The program prints the log Bayes factor of the odds ratio uniformity tests,
 * aggregated across the repetitions, for each ciphertext letter position and
 * each ciphertext digram position. If the <TT>-v</TT> flag is specified, the
 * program also prints the log Bayes factors for each repetition individually.
 *
 * @author  Alan Kaminsky
 * @version 23-Mar-2017
 */
public class StatTest01
	extends Task
	{

	// Command line arguments.
	boolean verbose;
	File dfreq;
	int Nlen;
	int Plen;
	int trials;
	int reps;
	long seed;

	// Global variables.
	DigramHist generatorHist;
	int allNonces;

	// Log Bayes factors for each repetition. First index is ciphertext
	// position. Second index is repetition.
	double[][] letterLbf;
	double[][] digramLbf;

	/**
	 * Main program.
	 */
	public void main
		(String[] args)
		throws Exception
		{
		// Parse command line arguments.
		int argi = 0;
		while (argi < args.length && args[argi].charAt(0) == '-')
			{
			switch (args[argi])
				{
				case "-v":
					verbose = true;
					break;
				default:
					usage();
				}
			++ argi;
			}
		if (args.length - argi != 6) usage();
		dfreq = new File (args[argi]);
		Nlen = Integer.parseInt (args[argi+1]);
		Plen = Integer.parseInt (args[argi+2]);
		trials = Integer.parseInt (args[argi+3]);
		reps = Integer.parseInt (args[argi+4]);
		seed = Long.parseLong (args[argi+5]);

		// Print provenance.
		System.out.printf ("$ java pj2 StatTest01");
		for (String arg : args) System.out.printf (" %s", arg);
		System.out.println();
		System.out.printf ("Nonce length = %d%n", Nlen);
		System.out.printf ("Plaintext length = %d%n", Plen);
		System.out.printf ("Trials per repetition = %d%n", trials);
		System.out.printf ("Repetitions = %d%n", reps);
		System.out.flush();

		// Read digram probabilities from file.
		InStream in = new InStream (new FileInputStream (dfreq));
		generatorHist = (DigramHist) in.readObject();
		in.close();

		// Decide whether to test all possible nonces.
		allNonces = 1;
		for (int i = 0; i < Nlen && allNonces <= trials; ++ i)
			allNonces *= 36;
		if (allNonces <= trials)
			System.out.printf ("Testing all %d nonces%n", allNonces);
		else
			System.out.printf ("Testing %d random nonces%n", trials);
		System.out.flush();

		// Set up storage for log Bayes factors.
		letterLbf = new double [Plen] [reps];
		digramLbf = new double [Plen-1] [reps];

		// Do repetitions in parallel.
		parallelFor (0, reps - 1) .schedule (dynamic) .exec (new Loop()
			{
			// Per-thread variables.
			int[] key;
			int[] nonce;
			int[] plaintext;
			int[] ciphertext;
			Histogram[] letterHist;
			Histogram[] digramHist;
			LC4 cipher;
			Random prng;
			RandomSubset rs;
			DigramPicker dp;

			// Initialize per-thread variables.
			public void start()
				{
				// Set up key, nonce, plaintext.
				key = new int [36];
				nonce = new int [Nlen];
				plaintext = new int [Plen];
				ciphertext = new int [Plen];

				// Set up histogram for each ciphertext letter.
				letterHist = new Histogram [Plen];
				for (int i = 0; i < Plen; ++ i)
					letterHist[i] = new Histogram (36);

				// Set up histogram for each ciphertext digram.
				digramHist = new Histogram [Plen-1];
				for (int i = 0; i < Plen - 1; ++ i)
					digramHist[i] = new Histogram (1296);

				// Set up cipher.
				cipher = new LC4();

				// Set up pseudorandom number generator.
				prng = new Random (seed + rank());
				rs = new RandomSubset (prng, 36, true);
				dp = new DigramPicker (prng, generatorHist);
				}

			// Perform one repetition.
			public void run (int rep)
				{
				// Clear histograms.
				for (int i = 0; i < Plen; ++ i)
					letterHist[i].clear();
				for (int i = 0; i < Plen - 1; ++ i)
					digramHist[i].clear();

				// Generate random key and plaintext.
				Util.genKey (key, rs);
				Util.genPlaintext (plaintext, dp);

				// Do trials for all possible nonces.
				if (allNonces <= trials)
					{
					for (int t = 0; t < allNonces; ++ t)
						{
						int tt = t;
						for (int i = 0; i < Nlen; ++ i)
							{
							nonce[i] = tt % 36;
							tt /= 36;
							}
						doTrial();
						}
					}

				// Do trials for random nonces.
				else
					{
					for (int t = 0; t < trials; ++ t)
						{
						Util.genNonce (nonce, prng);
						doTrial();
						}
					}

//for (int i = 0; i < 36; ++ i) System.out.printf ("%d\t%d%n", i, letterHist[0].count (i));

				// Compute log Bayes factors.
				for (int i = 0; i < Plen; ++ i)
					letterLbf[i][rep] = letterHist[i].logBayesFactor();
				for (int i = 0; i < Plen - 1; ++ i)
					digramLbf[i][rep] = digramHist[i].logBayesFactor();
				}

			// Do one trial.
			private void doTrial()
				{
				// Compute ciphertext.
				cipher.encrypt (key, nonce, plaintext, ciphertext);

				// Accumulate histograms of ciphertext letters.
				for (int i = 0; i < Plen; ++ i)
					{
					int c = ciphertext[i];
					letterHist[i].accumulate (c);
					}

				// Accumulate histograms of ciphertext digrams.
				for (int i = 0; i < Plen - 1; ++ i)
					{
					int d = ciphertext[i]*36 + ciphertext[i+1];
					digramHist[i].accumulate (d);
					}
				}
			});

		// Print log Bayes factors for each repetition.
		if (verbose)
			{
			printPerRepLbfs (letterLbf, "letter");
			printPerRepLbfs (digramLbf, "digram");
			}

		// Print aggregate log Bayes factors.
		printAggLbfs();
		}

	/**
	 * Print per-repetition log Bayes factors.
	 */
	private static void printPerRepLbfs
		(double[][] lbf,
		 String label)
		{
		int NR = lbf.length;
		int NC = lbf[0].length;
		Tabular table = new Tabular();
		int r;
		table.addColumns (1 + NC);
		r = table.addRow();
		table.set (r, 0, "Posn");
		for (int j = 0; j < NC; ++ j)
			table.setf (r, j + 1, "Rep_%d", j);
		for (int i = 0; i < NR; ++ i)
			{
			r = table.addRow();
			table.setf (r, 0, "%d", i);
			for (int j = 0; j < NC; ++ j)
				table.setf (r, j + 1, "%.5g", lbf[i][j]);
			}
		System.out.printf
			("======== Ciphertext %s per-rep log Bayes factors ========%n",
			 label);
		table.print();
		}

	/**
	 * Print aggregate log Bayes factors.
	 */
	private void printAggLbfs()
		{
		Tabular table = new Tabular();
		int r;
		table.addColumns (3);
		r = table.addRow();
		table.set (r, 1, "Letters");
		table.set (r, 2, "Digrams");
		r = table.addRow();
		table.set (r, 0, "Posn");
		table.set (r, 1, "log bf");
		table.set (r, 2, "log bf");
		for (int i = 0; i < Plen - 1; ++ i)
			{
			r = table.addRow();
			table.setf (r, 0, "%d", i);
			table.setf (r, 1, "%.5g", sum (letterLbf[i]));
			table.setf (r, 2, "%.5g", sum (digramLbf[i]));
			}
		r = table.addRow();
		table.setf (r, 0, "%d", Plen - 1);
		table.setf (r, 1, "%.5g", sum (letterLbf[Plen-1]));
		System.out.printf ("======== Aggregate log Bayes factors ========%n");
		table.print();
		}

	/**
	 * Returns the sum of the given array elements.
	 */
	private static double sum
		(double[] x)
		{
		double rv = 0;
		for (int i = 0; i < x.length; ++ i) rv += x[i];
		return rv;
		}

	/**
	 * Print a usage message and exit.
	 */
	private static void usage()
		{
		System.err.println ("Usage: java pj2 StatTest01 [-v] <dfreq> <Nlen> <Plen> <trials> <reps> <seed>");
		System.err.println ("-v = Print verbose output");
		System.err.println ("<dfreq> = File of digram frequencies");
		System.err.println ("<Nlen> = Nonce length");
		System.err.println ("<Plen> = Plaintext length");
		System.err.println ("<trials> = Number of trials per repetition");
		System.err.println ("<reps> = Number of repetitions");
		System.err.println ("<seed> = Random seed");
		System.exit (1);
		}

	}
