//******************************************************************************
//
// File:    StateRecovery02.java
// Package: ---
// Unit:    Class StateRecovery02
//
// This Java source file is copyright (C) 2017 by Alan Kaminsky. All rights
// reserved. For further information, contact the author, Alan Kaminsky, at
// ark@cs.rit.edu.
//
// This Java source file is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option) any
// later version.
//
// This Java source file is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// A copy of the GNU General Public License is provided in the file gpl.txt. You
// may also obtain a copy of the GNU General Public License on the World Wide
// Web at http://www.gnu.org/licenses/gpl.html.
//
//******************************************************************************

import edu.rit.io.InStream;
import edu.rit.pj2.Job;
import edu.rit.pj2.Task;
import edu.rit.pj2.tuple.ObjectTuple;
import edu.rit.util.Random;
import edu.rit.util.RandomSubset;
import java.io.File;
import java.io.FileInputStream;
import java.util.Arrays;

/**
 * Class StateRecovery02 is a parallel program that performs a series of state
 * recovery attacks on the LC4 cipher.
 * <P>
 * Usage: <TT>java pj2 StateRecovery02 <I>dfreq</I> <I>R</I> <I>K</I> <I>N</I>
 * <I>P</I> <I>seed</I></TT>
 * <P>
 * The program performs <I>R</I> repetitions in parallel. For each repetition,
 * the program picks a random key; a random nonce of length <I>Nlen</I>; and a
 * random plaintext of length <I>Plen</I> using the digram frequencies in the
 * <I>dfreq</I> file. The program computes the ciphertext. The program "forgets"
 * the final <I>K</I> letters of the key. The program does a branch-and-bound
 * search to recover the missing key letters and reports the number of conflicts
 * encountered during the search.
 *
 * @author  Alan Kaminsky
 * @version 29-Mar-2017
 */
public class StateRecovery02
	extends Job
	{
	private static final int N = 6;
	private static final int NSQR = N*N;

	/**
	 * Job main program.
	 */
	public void main
		(String[] args)
		throws Exception
		{
		// Parse command line arguments.
		if (args.length != 6) usage();
		File dfreq = new File (args[0]);
		int R = Integer.parseInt (args[1]);
		int K = Integer.parseInt (args[2]);
		int E = Integer.parseInt (args[3]);
		int P = Integer.parseInt (args[4]);
		long seed = Long.parseLong (args[5]);

		// Print provenance.
		System.out.printf ("$ java StateRecovery02");
		for (String arg : args) System.out.printf (" %s", arg);
		System.out.println();
		System.out.flush();

		// Read digram probabilities from file, put into tuple space.
		InStream in = new InStream (new FileInputStream (dfreq));
		DigramHist generatorHist = (DigramHist) in.readObject();
		in.close();
		putTuple (new ObjectTuple<DigramHist> (generatorHist));

		// Run each repetition as a separate task.
		for (int rep = 1; rep <= R; ++ rep)
			rule() .task (WorkerTask.class)
				.args (""+K, ""+E, ""+P, ""+seed, ""+rep);
		}

	/**
	 * Worker task.
	 */
	private static class WorkerTask
		extends Task
		{
		// Command line arguments.
		private int K;
		private int E;
		private int P;
		private long seed;
		private int rep;

		/**
		 * Worker task main program.
		 */
		public void main
			(String[] args)
			throws Exception
			{
			long t1 = System.currentTimeMillis();

			// Parse command line arguments.
			K = Integer.parseInt (args[0]);
			E = Integer.parseInt (args[1]);
			P = Integer.parseInt (args[2]);
			seed = Long.parseLong (args[3]);
			rep = Integer.parseInt (args[4]);

			// Read digram probabilities from tuple space.
			DigramHist generatorHist =
				readTuple (new ObjectTuple<DigramHist>()) .item;

			// Generate random key.
			Random prng = new Random (seed + rep);
			RandomSubset rs = new RandomSubset (prng, NSQR, true);
			int[] origKey = new int [NSQR];
			Util.genKey (origKey, rs);

			// Generate random plaintext and nonce.
			int M = E + P;
			int[] plaintext = new int [M];
			DigramPicker dp = new DigramPicker (prng, generatorHist);
			Util.genPlaintext (plaintext, dp);
			for (int i = 0; i < E; ++ i)
				plaintext[i] = prng.nextInt (NSQR);

			// Compute ciphertext.
			LC4 cipher = new LC4();
			int[] ciphertext = new int [M];
			cipher.encrypt (origKey, new int [0], plaintext, ciphertext);

			// Find state.
			State origState = new State (plaintext, ciphertext, origKey, K);
			State recovState = origState.verify();
			long t2 = System.currentTimeMillis();

			synchronized (System.out)
				{
				// Print inputs.
				System.out.printf ("======== Rep %d ========%n", rep);
				printLetters (origKey, "Key");
				printLetters (plaintext, "Plaintext ");
				printLetters (ciphertext, "Ciphertext");
				System.out.printf ("%d missing key letters%n", K);

				// Print results.
				if (recovState == null)
					System.out.printf ("Recovery failed!%n");
				else
					{
					int[] recovKey = recovState.getKey();
					printLetters (recovKey, "Recovered key");
					if (Arrays.equals (origKey, recovKey))
						System.out.printf ("Recovered key matches!%n");
					else
						System.out.printf ("Recovered key doesn't match!%n");
					}
				System.out.printf ("%d conflicts%n",
					recovState.conflictCount());
				System.out.printf ("%d msec%n", t2 - t1);
				System.out.flush();
				}
			}

		/**
		 * The worker task requires one core.
		 */
		protected static int coresRequired()
			{
			return 1;
			}
		}

	/**
	 * Print the given integer array as a string of letters.
	 */
	private static void printLetters
		(int[] s,
		 String label)
		{
		System.out.printf ("%s = %s%n", label, Util.intsToLetters (s));
		}

	/**
	 * Print a usage message and exit.
	 */
	private static void usage()
		{
		System.err.println ("Usage: java pj2 StateRecovery02 <dfreq> <R> <K> <N> <P> <seed>");
		System.err.println ("<dfreq> = File of digram frequencies");
		System.err.println ("<R> = Number of repetitions");
		System.err.println ("<K> = Number of missing key letters");
		System.err.println ("<N> = Nonce length");
		System.err.println ("<P> = Plaintext length");
		System.err.println ("<seed> = Random seed");
		System.exit (1);
		}

	}
